#silent
s       snippets
        a=`cat %b`
        if [ "$a" = "fori" ]; then
        echo "for (i = 0; i _; i++)" > %b
        fi
        if [ "$a" = "ife" ]; then
        cat <<EOF > %b
        if ( )
        {
        }
        else
        {
        }
        EOF
        fi
        if [ "$a" = "GPL" ]; then
        cat >>%b <<EOF
        /*
        *  This program is free software: you can redistribute it and/or modify
        *  it under the terms of the GNU General Public License as published by
        *  the Free Software Foundation, either version 3 of the License, or
        *  (at your option) any later version.
        *
        *  This program is distributed in the hope that it will be useful,
        *  but WITHOUT ANY WARRANTY; without even the implied warranty of
        *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        *  GNU General Public License for more details.
        *
        *  You should have received a copy of the GNU General Public License
        *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
        */
        EOF
        fi
        if [ "$a" = "type" ]; then
        cat <<EOF > %b
        typedef struct {
        ;
        } ?;
        EOF
        fi
